from typing import Any, Tuple, Dict, Iterator

import aiohttp
import ijson
from ijson import IncompleteJSONError

from azure.kusto.data._models import WellKnownDataSet
from azure.kusto.data.exceptions import KustoTokenParsingError, KustoUnsupportedApiError, KustoMultiApiError
from azure.kusto.data.streaming_response import JsonTokenType, FrameType, JsonToken


class JsonTokenReader:
    def __init__(self, stream: aiohttp.StreamReader):
        self.json_iter = ijson.parse_async(stream, use_float=True)

    def __aiter__(self) -> "JsonTokenReader":
        return self

    def __anext__(self) -> JsonToken:
        return self.read_next_token_or_throw()

    async def read_next_token_or_throw(self) -> JsonToken:
        try:
            next_item = await self.json_iter.__anext__()
        except IncompleteJSONError:
            next_item = None
        if next_item is None:
            raise KustoTokenParsingError("Unexpected end of stream")
        (token_path, token_type, token_value) = next_item

        return JsonToken(token_path, JsonTokenType[token_type.upper()], token_value)

    async def read_token_of_type(self, *token_types: JsonTokenType) -> JsonToken:
        token = await self.read_next_token_or_throw()
        if token.token_type not in token_types:
            raise KustoTokenParsingError(f"Expected one the following types: '{','.join(t.name for t in token_types)}' , got type {token.token_type}")
        return token

    async def read_start_object(self) -> JsonToken:
        return await self.read_token_of_type(JsonTokenType.START_MAP)

    async def read_start_array(self) -> JsonToken:
        return await self.read_token_of_type(JsonTokenType.START_ARRAY)

    async def read_string(self) -> str:
        return (await self.read_token_of_type(JsonTokenType.STRING)).token_value

    async def read_boolean(self) -> bool:
        return (await self.read_token_of_type(JsonTokenType.BOOLEAN)).token_value

    async def read_number(self) -> float:
        return (await self.read_token_of_type(JsonTokenType.NUMBER)).token_value

    async def skip_children(self, prev_token: JsonToken):
        if prev_token.token_type == JsonTokenType.MAP_KEY:
            prev_token = await self.read_next_token_or_throw()
        if prev_token.token_type in JsonTokenType.start_tokens():
            async for potential_end_token in self:
                if potential_end_token.token_path == prev_token.token_path and potential_end_token.token_type in JsonTokenType.end_tokens():
                    break

    async def skip_until_property_name(self, name: str) -> JsonToken:
        while True:
            token = await self.read_token_of_type(JsonTokenType.MAP_KEY)
            if token.token_value == name:
                return token

            await self.skip_children(token)

    async def skip_until_any_property_name(self, *names: str) -> JsonToken:
        while True:
            token = await self.read_token_of_type(JsonTokenType.MAP_KEY)
            if token.token_value in names:
                return token

            await self.skip_children(token)

    async def skip_until_property_name_or_end_object(self, *names: str) -> JsonToken:
        async for token in self:
            if token.token_type == JsonTokenType.END_MAP:
                return token

            if token.token_type == JsonTokenType.MAP_KEY:
                if token.token_value in names:
                    return token

                await self.skip_children(token)
                continue

            raise Exception(f"Unexpected token {token}")

    async def skip_until_token_with_paths(self, *tokens: (JsonTokenType, str)) -> JsonToken:
        async for token in self:
            if any((token.token_type == t_type and token.token_path == t_path) for (t_type, t_path) in tokens):
                return token
            await self.skip_children(token)


class StreamingDataSetEnumerator:
    def __init__(self, reader: JsonTokenReader):
        self.reader = reader
        self.done = False
        self.started = False
        self.started_primary_results = False
        self.finished_primary_results = False

    def __aiter__(self) -> "StreamingDataSetEnumerator":
        return self

    async def __anext__(self) -> Dict[str, Any]:
        if self.done:
            raise StopIteration()

        if not self.started:
            await self.reader.read_start_array()
            self.started = True

        token = await self.reader.skip_until_token_with_paths((JsonTokenType.START_MAP, "item"), (JsonTokenType.END_ARRAY, ""))
        if token == JsonTokenType.END_ARRAY:
            self.done = True
            raise StopIteration()

        frame_type = await self.read_frame_type()
        parsed_frame = await self.parse_frame(frame_type)
        is_primary_result = parsed_frame["FrameType"] == FrameType.DataTable and parsed_frame["TableKind"] == WellKnownDataSet.PrimaryResult.value
        if is_primary_result:
            self.started_primary_results = True
        elif self.started_primary_results:
            self.finished_primary_results = True

        return parsed_frame

    async def parse_frame(self, frame_type: FrameType) -> Dict[str, Any]:
        if frame_type == FrameType.DataSetHeader:
            frame = await self.extract_props(frame_type, ("IsProgressive", JsonTokenType.BOOLEAN), ("Version", JsonTokenType.STRING))
            if frame["IsProgressive"]:
                raise KustoUnsupportedApiError.progressive_api_unsupported()
            return frame
        if frame_type in [FrameType.TableHeader, FrameType.TableFragment, FrameType.TableCompletion, FrameType.TableProgress]:
            raise KustoUnsupportedApiError.progressive_api_unsupported()
        if frame_type == FrameType.DataTable:
            props = await self.extract_props(
                frame_type,
                ("TableId", JsonTokenType.NUMBER),
                ("TableKind", JsonTokenType.STRING),
                ("TableName", JsonTokenType.STRING),
                ("Columns", JsonTokenType.START_ARRAY),
            )
            await self.reader.skip_until_property_name("Rows")
            props["Rows"] = self.row_iterator()
            if props["TableKind"] != WellKnownDataSet.PrimaryResult.value:
                props["Rows"] = [r async for r in props["Rows"]]
            return props
        if frame_type == FrameType.DataSetCompletion:
            res = await self.extract_props(frame_type, ("HasErrors", JsonTokenType.BOOLEAN), ("Cancelled", JsonTokenType.BOOLEAN))
            token = await self.reader.skip_until_property_name_or_end_object("OneApiErrors")
            if token.token_type != JsonTokenType.END_MAP:
                res["OneApiErrors"] = self.parse_array(skip_start=False)
            return res

    async def row_iterator(self) -> Iterator[list]:
        await self.reader.read_token_of_type(JsonTokenType.START_ARRAY)
        while True:
            token = await self.reader.read_token_of_type(JsonTokenType.START_ARRAY, JsonTokenType.END_ARRAY, JsonTokenType.START_MAP)
            if token.token_type == JsonTokenType.START_MAP:
                raise KustoMultiApiError([await self.parse_object(skip_start=True)])
            if token.token_type == JsonTokenType.END_ARRAY:
                return
            yield await self.parse_array(skip_start=True)

    async def parse_array(self, skip_start: bool) -> list:
        if not skip_start:
            await self.reader.read_start_array()
        arr = []

        while True:
            token = await self.reader.read_token_of_type(
                JsonTokenType.NULL,
                JsonTokenType.BOOLEAN,
                JsonTokenType.NUMBER,
                JsonTokenType.STRING,
                JsonTokenType.START_MAP,
                JsonTokenType.START_ARRAY,
                JsonTokenType.END_ARRAY,
            )

            if token.token_type == JsonTokenType.END_ARRAY:
                return arr

            if token.token_type == JsonTokenType.START_MAP:
                arr.append(await self.parse_object(skip_start=True))
            elif token.token_type == JsonTokenType.START_ARRAY:
                arr.append(await self.parse_array(skip_start=True))
            else:
                arr.append(token.token_value)

    async def parse_object(self, skip_start: bool) -> Dict[str, Any]:
        if not skip_start:
            await self.reader.read_start_object()

        obj = {}
        while True:
            token_prop_name = await self.reader.read_token_of_type(JsonTokenType.MAP_KEY, JsonTokenType.END_MAP)
            if token_prop_name.token_type == JsonTokenType.END_MAP:
                return obj
            prop_name = token_prop_name.token_value

            token = await self.reader.read_token_of_type(
                JsonTokenType.NULL, JsonTokenType.BOOLEAN, JsonTokenType.NUMBER, JsonTokenType.STRING, JsonTokenType.START_MAP, JsonTokenType.START_ARRAY
            )

            if token.token_type == JsonTokenType.START_MAP:
                obj[prop_name] = await self.parse_object(skip_start=True)
            elif token.token_type == JsonTokenType.START_ARRAY:
                obj[prop_name] = await self.parse_array(skip_start=True)
            else:
                obj[prop_name] = token.token_value

    async def extract_props(self, frame_type: FrameType, *props: Tuple[str, JsonTokenType]) -> Dict[str, Any]:
        result = {"FrameType": frame_type}
        props_dict = dict(props)
        while props_dict:
            name = (await self.reader.skip_until_any_property_name(*props_dict.keys())).token_value
            if props_dict[name] == JsonTokenType.START_ARRAY:
                result[name] = await self.parse_array(skip_start=False)
            else:
                result[name] = (await self.reader.read_token_of_type(props_dict[name])).token_value
            props_dict.pop(name)

        return result

    async def read_frame_type(self) -> FrameType:
        await self.reader.skip_until_property_name("FrameType")
        return FrameType[await self.reader.read_string()]



import os
from dataclasses import dataclass, astuple
from typing import Optional


def get_env(*args, optional=False, default=None):
    """Return the first environment variable that is defined."""
    for arg in args:
        if arg in os.environ:
            return os.environ[arg]
    if optional or default:
        return default
    raise ValueError("No environment variables found: {}".format(args))


def set_env(key, value):
    """Set the environment variable."""
    os.environ[key] = value


def get_app_id(optional=False):
    """Return the app id."""
    result = get_env("APP_ID", "AZURE_CLIENT_ID", optional=optional)
    if result:
        set_env("AZURE_CLIENT_ID", result)
    return result


def get_auth_id(optional=False):
    """Return the auth id."""
    result = get_env("AUTH_ID", "APP_AUTH_ID", "AZURE_TENANT_ID", optional=optional)
    if result:
        set_env("AZURE_TENANT_ID", result)
    return result


def get_app_key(optional=False):
    """Return the app key."""
    result = get_env("APP_KEY", "AZURE_CLIENT_SECRET", optional=optional)
    if result:
        set_env("AZURE_CLIENT_SECRET", result)
    return result


@dataclass(frozen=True)
class AppKeyAuth:
    app_id: str
    app_key: str
    auth_id: str

    def __iter__(self):
        return iter(astuple(self))


def prepare_app_key_auth(optional=False) -> Optional[AppKeyAuth]:
    """Gets app key auth information from the env, sets the correct values for azidentity, and returns the AppKeyAuth object."""
    app_id = get_app_id(optional=optional)
    app_key = get_app_key(optional=optional)
    auth_id = get_auth_id(optional=optional)
    if app_id and app_key and auth_id:
        return AppKeyAuth(app_id, app_key, auth_id)
    return None



import functools
import os
import re
import sys
from dataclasses import dataclass
from typing import List, Tuple, Optional

from .env_utils import get_env
from azure.kusto.data._version import VERSION

NONE = "[none]"

REPLACE_REGEX = re.compile(r"[\r\n\s{}|]+")


@functools.lru_cache(maxsize=1)
def default_script() -> str:
    """Returns the name of the script that is currently running"""
    try:
        return os.path.basename(sys.argv[0]) or NONE
    except Exception:
        return NONE


@functools.lru_cache(maxsize=1)
def get_user_from_env() -> str:
    user = get_env("USERNAME", optional=True)
    domain = get_env("USERDOMAIN", optional=True)
    if domain and user:
        user = domain + "\\" + user
    if user:
        return user
    return NONE


@functools.lru_cache(maxsize=1)
def default_user():
    """Returns the name of the user that is currently logged in"""
    try:
        return os.getlogin() or get_user_from_env()
    except Exception:
        return get_user_from_env()


@functools.lru_cache(maxsize=1)
def format_version():
    return format_header(
        [
            ("Kusto.Python.Client", VERSION),
            (f"Runtime.{escape_field(sys.implementation.name)}", sys.version),
        ]
    )


def format_header(args: List[Tuple[str, str]]) -> str:
    return "|".join(f"{key}:{escape_field(val)}" for (key, val) in args if key and val)


def escape_field(field: str):
    return f"{{{REPLACE_REGEX.sub('_', field)}}}"


@dataclass
class ClientDetails:
    application_for_tracing: str
    user_name_for_tracing: str
    version_for_tracing: str = format_version()

    def __post_init__(self):
        self.application_for_tracing = self.application_for_tracing or default_script()
        self.user_name_for_tracing = self.user_name_for_tracing or default_user()

    @staticmethod
    def set_connector_details(
        name: str,
        version: str,
        app_name: Optional[str] = None,
        app_version: Optional[str] = None,
        send_user: bool = False,
        override_user: Optional[str] = None,
        additional_fields: Optional[List[Tuple[str, str]]] = None,
    ) -> "ClientDetails":
        params = [("Kusto." + name, version)]

        app_name = app_name or default_script()
        app_version = app_version or NONE

        params.append(("App." + escape_field(app_name), app_version))
        params.extend(additional_fields or [])

        user = NONE

        if send_user:
            user = override_user or default_user()

        return ClientDetails(application_for_tracing=format_header(params), user_name_for_tracing=user)



import abc
import io
import json
import uuid
from datetime import timedelta
from typing import Union, Optional, Any, NoReturn, ClassVar, TYPE_CHECKING
from urllib.parse import urljoin

from requests import Response, Session

from azure.kusto.data._cloud_settings import CloudSettings
from azure.kusto.data._token_providers import CloudInfoTokenProvider
from .client_details import ClientDetails
from .client_request_properties import ClientRequestProperties
from .exceptions import KustoServiceError, KustoThrottlingError, KustoApiError
from .kcsb import KustoConnectionStringBuilder
from .kusto_trusted_endpoints import well_known_kusto_endpoints
from .response import KustoResponseDataSet, KustoResponseDataSetV2, KustoResponseDataSetV1
from .security import _AadHelper

if TYPE_CHECKING:
    import aiohttp


class _KustoClientBase(abc.ABC):
    API_VERSION = "2024-12-12"

    _mgmt_default_timeout: ClassVar[timedelta] = timedelta(hours=1, seconds=30)
    _query_default_timeout: ClassVar[timedelta] = timedelta(minutes=4, seconds=30)
    _streaming_ingest_default_timeout: ClassVar[timedelta] = timedelta(minutes=10)
    _client_server_delta: ClassVar[timedelta] = timedelta(seconds=30)

    _aad_helper: _AadHelper
    client_details: ClientDetails
    _endpoint_validated = False
    _session: Union["aiohttp.ClientSession", "Session"]

    def __init__(self, kcsb: Union[KustoConnectionStringBuilder, str], is_async):
        self._kcsb = kcsb
        self._proxy_url: Optional[str] = None
        if not isinstance(kcsb, KustoConnectionStringBuilder):
            self._kcsb = KustoConnectionStringBuilder(kcsb)
        self._kusto_cluster = self._kcsb.data_source

        # notice that in this context, federated actually just stands for aad auth, not aad federated auth (legacy code)
        self._aad_helper = _AadHelper(self._kcsb, is_async) if self._kcsb.aad_federated_security else None

        if not self._kusto_cluster.endswith("/"):
            self._kusto_cluster += "/"

        # Create a session object for connection pooling
        self._mgmt_endpoint = urljoin(self._kusto_cluster, "v1/rest/mgmt")
        self._query_endpoint = urljoin(self._kusto_cluster, "v2/rest/query")
        self._streaming_ingest_endpoint = urljoin(self._kusto_cluster, "v1/rest/ingest/")
        self._request_headers = {
            "Accept": "application/json",
            "Accept-Encoding": "gzip,deflate",
            "x-ms-version": self.API_VERSION,
        }

        self.client_details = self._kcsb.client_details
        self._is_closed: bool = False

        self.default_database = self._kcsb.initial_catalog

    def _get_database_or_default(self, database_name: Optional[str]) -> str:
        return database_name or self.default_database

    def close(self):
        self._is_closed = True

    def set_proxy(self, proxy_url: str):
        self._proxy_url = proxy_url
        if self._aad_helper:
            self._aad_helper.token_provider.set_proxy(proxy_url)
            if isinstance(self._session, Session):
                self._aad_helper.token_provider.set_session(self._session)

    def validate_endpoint(self):
        if not self._endpoint_validated and self._aad_helper is not None:
            if isinstance(self._aad_helper.token_provider, CloudInfoTokenProvider):
                endpoint = CloudSettings.get_cloud_info_for_cluster(
                    self._kusto_cluster,
                    self._aad_helper.token_provider._proxy_dict,
                    self._session if isinstance(self._session, Session) else None,
                ).login_endpoint
                well_known_kusto_endpoints.validate_trusted_endpoint(
                    self._kusto_cluster,
                    endpoint,
                )
            self._endpoint_validated = True

    @staticmethod
    def _kusto_parse_by_endpoint(endpoint: str, response_json: Any) -> KustoResponseDataSet:
        if endpoint.endswith("v2/rest/query"):
            return KustoResponseDataSetV2(response_json)
        return KustoResponseDataSetV1(response_json)

    @staticmethod
    def _handle_http_error(
        exception: Exception,
        endpoint: Optional[str],
        payload: Optional[io.IOBase],
        response: "Union[Response, aiohttp.ClientResponse]",
        status: int,
        response_json: Any,
        response_text: Optional[str],
    ) -> NoReturn:
        if status == 404:
            if payload:
                raise KustoServiceError("The ingestion endpoint does not exist. Please enable streaming ingestion on your cluster.", response) from exception

            raise KustoServiceError(f"The requested endpoint '{endpoint}' does not exist.", response) from exception

        if status == 429:
            raise KustoThrottlingError("The request was throttled by the server.", response) from exception

        if status == 401:
            raise KustoServiceError("401. Missing adequate access rights.", response) from exception

        if payload:
            message = f"An error occurred while trying to ingest: Status: {status}, Reason: {response.reason}, Text: {response_text}."
            if response_json:
                raise KustoApiError(response_json, message, response) from exception

            raise KustoServiceError(message, response) from exception

        if response_json:
            raise KustoApiError(response_json, http_response=response) from exception

        if response_text:
            raise KustoServiceError(response_text, response) from exception

        raise KustoServiceError("Server error response contains no data.", response) from exception


class ExecuteRequestParams:
    @staticmethod
    def _from_stream(
        stream: io.IOBase,
        properties: ClientRequestProperties,
        request_headers: Any,
        timeout: timedelta,
        mgmt_default_timeout: timedelta,
        client_server_delta: timedelta,
        client_details: ClientDetails,
    ):
        # Before 3.0 it was KPC.execute_streaming_ingest, but was changed to align with the other SDKs
        client_request_id_prefix = "KPC.executeStreamingIngest;"
        request_headers = request_headers.copy()
        request_headers["Content-Encoding"] = "gzip"
        if properties:
            request_headers.update(json.loads(properties.to_json())["Options"])

        return ExecuteRequestParams(
            stream, None, request_headers, client_request_id_prefix, properties, timeout, mgmt_default_timeout, client_server_delta, client_details
        )

    @staticmethod
    def _from_query(
        query: str,
        database: str,
        properties: ClientRequestProperties,
        request_headers: Any,
        timeout: timedelta,
        mgmt_default_timeout: timedelta,
        client_server_delta: timedelta,
        client_details: ClientDetails,
    ):
        json_payload = {"db": database, "csl": query}
        if properties:
            json_payload["properties"] = properties.to_json()

        client_request_id_prefix = "KPC.execute;"
        request_headers = request_headers.copy()
        request_headers["Content-Type"] = "application/json; charset=utf-8"

        return ExecuteRequestParams(
            None, json_payload, request_headers, client_request_id_prefix, properties, timeout, mgmt_default_timeout, client_server_delta, client_details
        )

    @staticmethod
    def _from_blob_url(
        blob: str,
        properties: ClientRequestProperties,
        request_headers: Any,
        timeout: timedelta,
        mgmt_default_timeout: timedelta,
        client_server_delta: timedelta,
        client_details: ClientDetails,
    ):
        json_payload = {"sourceUri": blob}
        client_request_id_prefix = "KPC.executeStreamingIngestFromBlob;"
        request_headers = request_headers.copy()
        request_headers["Content-Type"] = "application/json; charset=utf-8"
        if properties:
            request_headers.update(json.loads(properties.to_json())["Options"])
        return ExecuteRequestParams(
            None, json_payload, request_headers, client_request_id_prefix, properties, timeout, mgmt_default_timeout, client_server_delta, client_details
        )

    def __init__(
        self,
        payload,
        json_payload,
        request_headers,
        client_request_id_prefix,
        properties: ClientRequestProperties,
        timeout: timedelta,
        mgmt_default_timeout: timedelta,
        client_server_delta: timedelta,
        client_details: ClientDetails,
    ):
        special_headers = [
            {
                "name": "x-ms-client-request-id",
                "value": client_request_id_prefix + str(uuid.uuid4()),
                "property": lambda p: p.client_request_id,
            },
            {
                "name": "x-ms-client-version",
                "value": client_details.version_for_tracing,
                "property": lambda p: None,
            },
            {
                "name": "x-ms-app",
                "value": client_details.application_for_tracing,
                "property": lambda p: p.application,
            },
            {
                "name": "x-ms-user",
                "value": client_details.user_name_for_tracing,
                "property": lambda p: p.user,
            },
        ]

        for header in special_headers:
            value: str
            if properties and header["property"](properties) is not None:
                value = header["property"](properties)
            else:
                value = header["value"]

            if value is not None:
                # Replace any characters that aren't ascii with '?'
                value = value.encode("ascii", "replace").decode("ascii", "strict")
                request_headers[header["name"]] = value

        if properties is not None:
            if properties.get_option(ClientRequestProperties.no_request_timeout_option_name, False):
                timeout = mgmt_default_timeout
            else:
                timeout = properties.get_option(ClientRequestProperties.request_timeout_option_name, timeout)

        timeout = (timeout or mgmt_default_timeout) + client_server_delta

        self.json_payload = json_payload
        self.request_headers = request_headers
        self.timeout = timeout
        self.payload = payload



from typing import AsyncIterator

from azure.kusto.data._models import KustoResultRow, BaseStreamingKustoResultTable


class KustoStreamingResultTable(BaseStreamingKustoResultTable):
    """Async Iterator over a Kusto result table."""

    async def __anext__(self) -> KustoResultRow:
        try:
            row = await self.raw_rows.__anext__()
        except StopAsyncIteration:
            self.finished = True
            raise
        self.row_count += 1
        return KustoResultRow(self.columns, row)

    def __aiter__(self) -> AsyncIterator[KustoResultRow]:
        return self



def aio_documented_by(original):
    def wrapper(target):
        target.__doc__ = "Aio function: {original_doc}".format(original_doc=original.__doc__)
        return target

    return wrapper


def documented_by(original):
    def wrapper(target):
        target.__doc__ = original.__doc__
        return target

    return wrapper



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from ._version import VERSION as __version__
from .client import KustoClient
from .client_request_properties import ClientRequestProperties
from .kcsb import KustoConnectionStringBuilder
from .data_format import DataFormat

__all__ = [
    "__version__",
    "KustoClient",
    "ClientRequestProperties",
    "KustoConnectionStringBuilder",
    "DataFormat",
]
